% [F matches] = BEEM2(p1_m,p2_m,rr1,rr2,distanceRatio)
%
% This function gets:
% p1_m, p2_m - putative match 
% rr1, rr2 -  SIFT parameters for the putative match. rr1 for p1_m and rr2 for p2_m. 
% The first column in rr1/rr2 is the rotation angle of the dominant gradient of the SIFT descriptor.
% The second column in rr1/rr2 is the scale in the space scale of the SIFT descriptor. 
% distanceRatio - the distance ratio of the closest to the second-closest
% it calculates F and supportted inliers and saves those to files

function  [F matches] = BEEM2(p1_m,p2_m,rr1,rr2,distanceRatio)
N = size(p1_m,1);
sprintf(' Number of putative matches is %d\n',N);
if(size(p1_m,2) ~= 2 )
    error('p1_m should have two columns');
end;
if(size(p2_m,2) ~= 2 )
    error('p2_m should have two columns');
end;
if(size(p2_m,1) ~= N )
    error('p2_m should have %d rows',N);
end;
if(size(rr1,2) ~= 2 )
    error('rr1 should have two columns');
end;
if(size(rr1,1) ~= N )
    error('rr1 should have %d rows',N);
end;
if(size(rr2,2) ~= 2 )
    error('rr2 should have two columns');
end;
if(size(rr2,1) ~= N )
    error('rr2 should have %d rows',N);
end;
if(size(distanceRatio,2) ~= 1 )
    error('distanceRatio should have one column');
end;
if(size(distanceRatio,1) ~= N )
    error('distanceRatio should have %d rows',N);
end;
save_file('p1_m',p1_m);
save_file('p2_m',p2_m);
save_file('rr1',rr1);
save_file('rr2',rr2);
save_file('distanceRatio',distanceRatio);
system('BEEM 2');
F = textread('F.txt');
F(:,4) = [];
matches = textread('matches.txt');
matches(:,5) = [];

function save_file(filename,A)
fid = fopen(filename, 'wt');
fprintf(fid, '%d\n%d\n', size(A));
fprintf(fid, '%6.2f\n', A);
fclose(fid);