#include "StdAfx.h"
#include "SiftUtil.h"
#include "MyMath.h"

void SiftUtil::readKeypoints2(char *filename , char *descOutputFile , char *corOutputFile){
	ofstream descOutFile(descOutputFile);
	if (! descOutFile)
		cerr << "Could not open file: " << descOutputFile;

	ofstream corOutFile(corOutputFile);
	if (! corOutFile)
		cerr << "Could not open file: " << corOutputFile;

	ifstream file(filename);
    if (! file)
		cerr << "Could not open file: " << filename;

    int  j, num, len;
	
	file >> num;
	file >> len;
	if (len != 128)
		cerr << "Keypoint descriptor length invalid (should be 128).";

	descOutFile << num << endl;
	descOutFile << len << endl;
	
	corOutFile << num << endl;
	corOutFile << 4 << endl;

	double** locs = 0;
	locs = new double*[num];
	int** descriptors = 0;
	descriptors = new int*[num];
	for (int k=0 ; k<num ; k++){
		locs[k] = new double[4];
		descriptors[k] = new int[len];
	}
	for (int k=0 ; k<num ; k++){
		// reading row, col, scale, orientation
		file >> locs[k][0];
		file >> locs[k][1];
		file >> locs[k][2];
		file >> locs[k][3];
	
		for (j = 0; j < len; j++) {
			file >> descriptors[k][j];
		}
		
	}
	
	//writing the correspondences to the file (row(x),col(y),scale,orientation)
	for (int w=0 ; w<4 ; w++){
		for (int k=0 ; k<num ; k++){
			corOutFile << locs[k][w] << endl;
		}
	}

	// write descriptors to file
	for (int ww=0 ; ww<len ; ww++){
		for (int kk = 0; kk < num; kk++) {
			descOutFile << descriptors[kk][ww] << endl;
		}
	}
	/*// write descriptors to file
	for (int ww=0 ; ww<len ; ww++){
		for (int kk = 0; kk < num; kk++) {
			descOutFile << descriptors[kk][ww] << " ";
		}
		descOutFile << endl;
	}*/
}

void SiftUtil::readKeypoints(char const *filename , MyMath::MyData& desc 
, MyMath::MyData& cor){
	MyMath myMathHelper1;
	ifstream file(filename);
	if (! file){
		cerr << "Could not open file: " << filename;
		exit(1);
	}

    int  j, num, len;

	file >> num;
	file >> len;
	if (len != 128)
		cerr << "Keypoint descriptor length invalid (should be 128).";

	desc = myMathHelper1.zeros(num,len);
	cor = myMathHelper1.zeros(num,4);
	
	double** locs = 0;
	locs = new double*[num];
	int** descriptors = 0;
	descriptors = new int*[num];
	for (int k=0 ; k<num ; k++){
		locs[k] = new double[4];
		descriptors[k] = new int[len];
	}

	for (int k=0 ; k<num ; k++){
		// reading row, col, scale, orientation
		file >> cor.getCell(k,0);
		file >> cor.getCell(k,1);
		file >> cor.getCell(k,2);
		file >> cor.getCell(k,3);
		
		for (j = 0; j < len; j++) {
			file >> desc.getCell(k,j);
		}
		
	}
}


void SiftUtil::myReadKeypoints(char *filename , char *descOutputFile , char *locOutputFile){
	ofstream descOutFile(descOutputFile);
	if (! descOutFile)
		cerr << "Could not open file: " << descOutputFile;

	ofstream locOutFile(locOutputFile);
	if (! locOutFile)
		cerr << "Could not open file: " << locOutputFile;

	ifstream file(filename);
    if (! file)
		cerr << "Could not open file: " << filename;

    int  j, num, len;
	
	file >> num;
	file >> len;
	if (len != 128)
		cerr << "Keypoint descriptor length invalid (should be 128).";

	descOutFile << num << endl;
	descOutFile << len << endl;

	locOutFile << num << endl;
	locOutFile << 4 << endl;

	double** locs = 0;
	locs = new double*[num];
	int** descriptors = 0;
	descriptors = new int*[num];
	for (int k=0 ; k<num ; k++){
		locs[k] = new double[4];
		descriptors[k] = new int[len];
	}

	double descTmp = 0;
	for (int k=0 ; k<num ; k++){
		// reading row, col, scale, orientation
		file >> locs[k][0];
		locOutFile << locs[k][0] << endl;
		file >> locs[k][1];
		locOutFile << locs[k][1] << endl;
		file >> locs[k][2];
		locOutFile << locs[k][2] << endl;
		file >> locs[k][3];
		locOutFile << locs[k][3] << endl;
	
		descTmp = 0;
		for (j = 0; j < len; j++) {
			file >> descriptors[k][j];
			descTmp += (descriptors[k][j] * descriptors[k][j]);
		}
		descTmp = sqrt(descTmp);
		for (j = 0; j < len; j++) {
			descOutFile << descriptors[k][j]/descTmp << endl;
		}

	}
}

void SiftUtil::myMatch(char *desc1File , char *loc1File, char *desc2File , char *loc2File , double dr){
	MyMath myMathHelper1;
	MyMath::MyData desc1,loc1,desc2,loc2,desc2tag;
	MyMath::MyData dotProds,vals;
	MyMath::MyData p1_m,p2_m,rr1,rr2,d_r;
	int indx[2];

	desc1 = myMathHelper1.createMat("desc1");
	loc1 = myMathHelper1.createMat("loc1");
	desc2 = myMathHelper1.createMat("desc2");
	loc2 = myMathHelper1.createMat("loc2");
	
	p1_m = myMathHelper1.zeros(desc1.numOfRows,2);
	p2_m = myMathHelper1.zeros(desc1.numOfRows,2);
	rr1 = myMathHelper1.zeros(desc1.numOfRows,2);
	rr2 = myMathHelper1.zeros(desc1.numOfRows,2);
	d_r = myMathHelper1.zeros(desc1.numOfRows,1);

	int j=0;
	desc2tag = myMathHelper1.transposeMatrix(desc2);
	for (int i = 0 ; i < desc1.numOfRows ; i++ ){
		if (i==1000 || i==3000 || i==6000 || i==8000 || i==10000 || i == 11000 || i == 12000 || i ==13000 || i ==14000)
		{
			cout << "i is: " << i << endl;
		}
		dotProds = myMathHelper1.mulMatrices( myMathHelper1.getRow(desc1,i) , desc2tag ); // Computes vector of dot products
		//[vals,indx] = sort(acos(dotProds));  % Take inverse cosine and sort results
		myMathHelper1.doAcos(dotProds);
		vals = myMathHelper1.sortOneRow(dotProds);
		indx[0] = myMathHelper1.findIndxInRow(dotProds,vals.getCell(0,0));
		indx[1] = myMathHelper1.findIndxInRow(dotProds,vals.getCell(0,1));

		// Check if nearest neighbor has angle less than distRatio times 2nd.
		if (vals.getCell(0,0) < dr * vals.getCell(0,1)){
			d_r.getCell(j,0) = vals.getCell(0,0)/vals.getCell(0,1);
			
			p1_m.getCell(j,0) = loc1.getCell(i,1);
			p1_m.getCell(j,1) = loc1.getCell(i,0);
			p2_m.getCell(j,0) = loc2.getCell(indx[0],1);
			p2_m.getCell(j,1) = loc2.getCell(indx[0],0);
			//rr1(j,:) = loc1(i,3:4);
			rr1.getCell(j,0) = loc1.getCell(i,2);
			rr1.getCell(j,1) = loc1.getCell(i,3);
			//rr2(j,:) = loc2(indx[1],3:4);
			rr2.getCell(j,0) = loc2.getCell(indx[0],2);
			rr2.getCell(j,1) = loc2.getCell(indx[0],3);
			j++;
		}
	}
	p1_m.setNumOfRows(j);
	p2_m.setNumOfRows(j);
	rr1.setNumOfRows(j);
	rr2.setNumOfRows(j);
	d_r.setNumOfRows(j);

	// create files : p1_m,p2_m,rr1,rr2,distanceRatio
	myMathHelper1.printMatToFile(p1_m,"p1_m");
	myMathHelper1.printMatToFile(p2_m,"p2_m");
	myMathHelper1.printMatToFile(rr1,"rr1");
	myMathHelper1.printMatToFile(rr2,"rr2");
	myMathHelper1.printMatToFile(d_r,"distanceRatio");

   /*dotprods = des1(i,:) * des2t;        % Computes vector of dot products
   [vals,indx] = sort(acos(dotprods));  % Take inverse cosine and sort results

   % Check if nearest neighbor has angle less than distRatio times 2nd.
   if (vals(1) < distRatio * vals(2))
       d_r(j,1) = vals(1)/vals(2);
       match(i) = indx(1);
       p1_m(j,1) = loc1(i,2);
       p1_m(j,2) = loc1(i,1);
       p2_m(j,1) = loc2(indx(1),2);
       p2_m(j,2) = loc2(indx(1),1);
       rr1(j,:) = loc1(i,3:4);
       rr2(j,:) = loc2(indx(1),3:4);
       j = j + 1;
   else
      match(i) = 0;
   end
end*/
}

SiftUtil::SiftUtil(void)
{
}

SiftUtil::~SiftUtil(void)
{
}
