#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#define NR_END 1
#define FREE_ARG char*

#define TINY (float)1.0e-20;

#define ITMAX 100
#define EPS 3.0e-7
#define FPMIN 1.0e-30


static double dsqrarg;
#define DSQR(a) ((dsqrarg=(a)) == 0.0 ? 0.0 : dsqrarg*dsqrarg)

static double dmaxarg1,dmaxarg2;
#define DMAX(a,b) (dmaxarg1=(a),dmaxarg2=(b),(dmaxarg1) > (dmaxarg2) ?\
        (dmaxarg1) : (dmaxarg2))

static int iminarg1,iminarg2;
#define IMIN(a,b) (iminarg1=(a),iminarg2=(b),(iminarg1) < (iminarg2) ?\
        (iminarg1) : (iminarg2))

#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))


class itayUtil{
public:
	itayUtil(void);
public:
	void nrerror(const char error_text[]);
	float *vector(long nl, long nh);
	int *ivector(long nl, long nh);
	unsigned char *cvector(long nl, long nh);
	unsigned long *lvector(long nl, long nh);
	double *dvector(long nl, long nh);
	float **matrix(long nrl, long nrh, long ncl, long nch);
	double **dmatrix(long nrl, long nrh, long ncl, long nch);
	int **imatrix(long nrl, long nrh, long ncl, long nch);
	float **submatrix(float **a, long oldrl, long oldrh, long oldcl, long oldch, long newrl, long newcl);
	float **convert_matrix(float *a, long nrl, long nrh, long ncl, long nch);
	float ***f3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh);
	void free_vector(float *v, long nl, long nh);
	void free_ivector(int *v, long nl, long nh);
	void free_cvector(unsigned char *v, long nl, long nh);
	void free_lvector(unsigned long *v, long nl, long nh);
	void free_dvector(double *v, long nl, long nh);
	void free_matrix(float **m, long nrl, long nrh, long ncl, long nch);
	void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch);
	void free_imatrix(int **m, long nrl, long nrh, long ncl, long nch);
	void free_submatrix(float **b, long nrl, long nrh, long ncl, long nch);
	void free_convert_matrix(float **b, long nrl, long nrh, long ncl, long nch);
	void free_f3tensor(float ***t, long nrl, long nrh, long ncl, long nch, long ndl, long ndh);
	
	/////
	void dsvdcmp(double **a, int m, int n, double w[], double **v);
	double dpythag(double a, double b);
	void ludcmp(float **a, int n, int *indx, float *d);
	void lubksb(float **a, int n, int *indx, float b[]);
	void dludcmp(double **a, int n, int *indx, double *d);
	void dlubksb(double **a, int n, int *indx, double b[]);

	////
	int quintic(double [], double [], double [], int*, double);
	int quartic(double[], double[], double[], int* );
	int cubic(double[], double[], int*);
	int signR(double);
	double CBRT(double);
	
	////
	double binomial_pdf ( int x, int a, double b );
	int binomial_coef ( int n, int k );
	int i4_max ( int i1, int i2 );
	int i4_min ( int i1, int i2 );

	////
	void fit(double x[], double y[], int ndata, double sig[], int mwt, double *a,
			double *b, double *siga, double *sigb, double *chi2, double *q);
	double gammq(double a, double x);
	void gcf(double *gammcf, double a, double x, double *gln);
	double gammln(double xx);
	void gser(double *gamser, double a, double x, double *gln);

	};
